// -*- mode:c++ -*-
//
// Module SmartRobot.cc
//
// Copyright (c) 2002 Raoul M. Gough
//
// History
// =======
// 2002/ 9/ 2	rmg	File creation
//
// $Id: SmartRobot.cc,v 1.5 2002/09/02 18:23:10 raoul Exp $
//

#include "SmartRobot.hh"
#include "World.hh"
#include "AddressHeap.hh"

#include <iostream>
#include <string>
#include <stdlib.h>
#include <stdexcept>
#include <boost/smart_ptr.hpp>
#include <sstream>

////////////////////////////////////////////////////////////////////////////
// Constructor
////////////////////////////////////////////////////////////////////////////

SmartRobot::SmartRobot (World &w, RobotInfo const &info)
 : Robot (w, info)
  , mGoal (-1)
  , mCommand ("Drop") // Safe no-op
  , mPackageChange (true)
  , mBaseEmergency (0)
{
}

////////////////////////////////////////////////////////////////////////////
// Decide the next move and go ahead and do it
////////////////////////////////////////////////////////////////////////////

void SmartRobot::makeMove (std::ostream &strm)
{
  mCommand = "Drop"; // Safe no-op

  if (location() == mGoal)
    {
      logStream() << "At previous goal\n";
      mGoal = -1;
    }

  if (deliveriesHere () || pickupsHere ())
    {
    }

  else if (moveToGoal ())
    {
    }

  else
    {
      std::cerr << ".";
      randomMove ();
    }

  sendCommand (strm, mCommand);
}

////////////////////////////////////////////////////////////////////////////
// Find the next delivery address
////////////////////////////////////////////////////////////////////////////

void SmartRobot::calcNextDelivery ()
{
  int cap = capacity ();

  AddressSet const &byAddress = packages().byAddress ();

  HeapType bestAddresses;

  resetHeap (bestAddresses
	     , byAddress
	     , *worldPtr()->tree (location())
	     , worldPtr()->board()
	     , cap);

  if (bestAddresses.size())
    {
      setGoal (bestAddresses.front().mLocation);
    }
}

////////////////////////////////////////////////////////////////////////////
// Make a move to the current goal
////////////////////////////////////////////////////////////////////////////

bool SmartRobot::moveToGoal ()
{
  if (mPackageChange)
    {
      // Load has changed - get next delivery address
      calcNextDelivery ();
      mPackageChange = false;
    }

  if (mGoal == -1)
    {
      // No delivery to do - find a pickup point
      calcNextPickup ();
    }

  int nextVertex = worldPtr()->tree (mGoal)->nextToRoot (location());

  int offset = nextVertex - location();

  debugStream() << "Offset to " << nextVertex << " = " << offset << "\n";

  if (offset == worldPtr()->board().width())
    {
      mCommand = "Move N";
    }

  else if (offset == -worldPtr()->board().width())
    {
      mCommand = "Move S";
    }

  else if (offset == 1)
    {
      mCommand = "Move E";
    }

  else if (offset == -1)
    {
      mCommand = "Move W";
    }

  else
    {
      return false;
    }

  return true;
}

////////////////////////////////////////////////////////////////////////////
// Check for deliveries at this square
////////////////////////////////////////////////////////////////////////////

bool SmartRobot::deliveriesHere ()
{
  Location loc = worldPtr()->board().whereIs (location());

  AddressSet::const_iterator iter = packages().beginAt (loc);
  AddressSet::const_iterator end = packages().endAt (loc);

  if (iter != end)
    {
      std::ostringstream dropCommand;

      dropCommand << "Drop";

      while (iter != end)
	{
	  dropCommand << " " << iter->mId;
	  ++iter;
	}

      mCommand = dropCommand.str();

      return true;
    }

  return false;
}

////////////////////////////////////////////////////////////////////////////
// Check for pickups at this square
////////////////////////////////////////////////////////////////////////////

bool SmartRobot::pickupsHere ()
{
  SquareInfo const *squarePtr = worldPtr()->knownInfo (location());

  bool result = false; // Assume failure

  if (squarePtr
      && squarePtr->packages().totalWeight()
      && (packages().totalWeight() < capacity()))
    {
      // We could decide to drop stuff first, but that we be hard
      int cap = capacity () - packages().totalWeight();

      AddressSet const &byAddress = squarePtr->packages().byAddress ();

      HeapType bestAddresses;

      resetHeap (bestAddresses
		 , byAddress
		 , *worldPtr()->tree (location())
		 , worldPtr()->board()
		 , cap);

      if (logStream().mStrmPtr)
	{
	  logStream()
	    << bestAddresses.size()
	    << " addresses: ";

	  for (int count = 0; count < bestAddresses.size(); ++count)
	    {
	      logStream() << bestAddresses[count] << " ";
	    }

	  logStream() << std::endl;
	}

      std::ostringstream pickCommand;

      pickCommand << "Pick";

      while ((cap >= 0) && bestAddresses.size())
	{
	  logStream () << "Trying address " << bestAddresses.front() << "\n";

	  Location loc
	    = worldPtr()->board().whereIs (bestAddresses.front().mLocation);

	  AddressSet::const_iterator iter
	    = squarePtr->packages().beginAt (loc);

	  AddressSet::const_iterator end
	    = squarePtr->packages().endAt (loc);

	  while ((iter != end) && (iter->mWeight <= cap))
	    {
	      result = true;

	      pickCommand << " " << iter->mId;
	      cap -= iter->mWeight;

	      ++iter;
	    }

	  while ((!bestAddresses.empty())
		 && (bestAddresses.back().metric() == 0))
	    {
	      bestAddresses.pop_back();
	    }

	  if (!bestAddresses.empty())
	    {
	      bestAddresses.erase (bestAddresses.begin());
	      recalcHeap (bestAddresses);
	    }
	}

      if (result)
	{
	  mCommand = pickCommand.str();
	}
    }

  return result;
}

////////////////////////////////////////////////////////////////////////////
// Make a random move
////////////////////////////////////////////////////////////////////////////

void SmartRobot::randomMove ()
{
  if ((random() % 100) < 40)
    {
      mCommand = "Drop"; // Do nothing
    }

  else
    {
      mCommand = std::string ("Move ") + "NSEW"[random() % 4];
    }
}

////////////////////////////////////////////////////////////////////////////
// Set our goal square
////////////////////////////////////////////////////////////////////////////

void SmartRobot::setGoal (int square)
{
  mGoal = square;

  logStream()
    << "Goal is "
    << square
    << " "
    << worldPtr()->board().whereIs (square)
    << "\n";

  if (debugStream().mStrmPtr)
    {
      Path path (worldPtr()->tree (mGoal)->pathFromRoot (location()));

      int start = location();

      debugStream()
	<< "Route from "
	<< location()
	<< " "
	<< worldPtr()->board().whereIs (location())
	<< " to "
	<< worldPtr()->board().whereIs (mGoal)
	<< ":";

      while (start != mGoal)
	{
	  int next = path.next (start, mGoal);
	  debugStream()
	    << " "
	    << next
	    << worldPtr()->board().whereIs (next);

	  start = next;
	}
      debugStream() << "\n";
    }
}

////////////////////////////////////////////////////////////////////////////
// Handle package load changes
////////////////////////////////////////////////////////////////////////////

void SmartRobot::pickPackage (PackageInfo const &package)
{
  mPackageChange = true;
  Robot::pickPackage (package);
}

PackageInfo SmartRobot::dropPackage (int id)
{
  mPackageChange = true;
  return Robot::dropPackage (id);
}

////////////////////////////////////////////////////////////////////////////
// Command senders
////////////////////////////////////////////////////////////////////////////

void SmartRobot::sendCommand (std::ostream &strm, char const *str)
{
  commStream() << "Client: " << "1 " << str << std::endl;

  strm << "1 " << str << std::endl;
}

void SmartRobot::sendCommand (std::ostream &strm, std::string const &str)
{
  sendCommand (strm, str.c_str());
}

////////////////////////////////////////////////////////////////////////////
// Calculate the next pickup (best package group or unexplored base)
////////////////////////////////////////////////////////////////////////////

void SmartRobot::calcNextPickup ()
{
  boost::shared_ptr<SpanningTree const>
    treePtr (worldPtr()->tree (location()));

  std::vector<AddressInfo>
    dests (worldPtr()->packagesAndBases());

  int cap = capacity();

  for (std::vector<AddressInfo>::iterator iter = dests.begin()
	 ; iter != dests.end()
	 ; ++iter)
    {
      iter->mCapacityPtr = &cap;
      iter->mDistance = treePtr->distance (iter->mLocation);
    }

  recalcHeap (dests);

  if (dests.empty())
    {
      setGoal (worldPtr()->board().bases()[mBaseEmergency]);

      mBaseEmergency
	= (mBaseEmergency + 1) % worldPtr()->board().bases().size();
    }

  else
    {
      setGoal (dests.front().mLocation);
    }
}
