// -*- mode:c++ -*-
//
// Module StaticBoard.cc
//
// Copyright (c) 2002 Raoul M. Gough
//
// History
// =======
// 2002/ 8/31	rmg	File creation
//
// $Id: StaticBoard.cc,v 1.9 2002/09/02 18:23:54 raoul Exp $
//

#include "StaticBoard.hh"

#include <iostream>
#include <string>
#include <stdexcept>
#include <algorithm>

#include <assert.h>

/////////////////////////////////////////////////////////////////////////
// Get original character representation
/////////////////////////////////////////////////////////////////////////

namespace
{
  const char waterChar = '~';
  const char baseChar = '@';
  const char wallChar = '#';
  const char freeChar = '.';

  void check (StaticBoard::BitSet b)
  {
    if ((b & SBUtils::flag_wall) && !SBUtils::isWall (b))
      {
	std::cerr << "!" << std::endl;
      }
  }
}

char StaticBoard::operator() (int x, int y) const
{
  BitSet const square = rawBits (offset (x, y));

 if (SBUtils::isLethal (square))
    {
      return waterChar;
    }

  else if (SBUtils::isBase (square))
    {
      return baseChar;
    }

  else if (SBUtils::isWall (square))
    {
      return wallChar;
    }

  else
    {
      return freeChar;
    }
}

/////////////////////////////////////////////////////////////////////////
// Load board from a stream
/////////////////////////////////////////////////////////////////////////

void StaticBoard::load (std::istream &strm)
{
  StaticBoard temp;

  strm >> temp.mWidth;
  strm >> temp.mHeight;

  if (strm)
    {
      temp.mBoardBits.resize (temp.mWidth * temp.mHeight);

      std::string prevRow;

      for (int count = 0; count < temp.mWidth; ++count)
	{
	  prevRow += wallChar;
	}

      std::string thisRow;

      for (int row = 0; row < temp.mHeight; ++row)
	{
	  strm >> thisRow;

	  if (thisRow.length() != temp.mWidth)
	    {
	      throw std::runtime_error
		(thisRow + ": invalid StaticBoard width");
	    }

	  int rawOffset = temp.offset (0, row);

	  for (int column = 0; column < thisRow.length(); ++column)
	    {
	      BitSet square (0);

	      switch (thisRow[column])
		{
		case wallChar:
		  square |= SBUtils::flag_wall;
		  // This square is uninteresting
		  break;

		case waterChar:
		  square |= SBUtils::flag_lethal;

		  if (column > 0)
		    {
		      temp.mBoardBits[rawOffset - 1]
			|= SBUtils::flag_lethal_neighbour;
		    }

		  check (temp.mBoardBits[rawOffset - 1]);

		  if (prevRow[column] == freeChar)
		    {
		      temp.mBoardBits[rawOffset - temp.mWidth]
			|= SBUtils::flag_lethal_neighbour;
		    }

		  check (temp.mBoardBits[rawOffset - temp.mWidth]);
		  break;

		case baseChar:
		  mBases.push_back (rawOffset);
		  square |= SBUtils::flag_base;
		  // FALL THROUGH

		case freeChar:
		  if (column > 0)
		    {
		      switch (thisRow[column - 1])
			{
			case wallChar:
			  // This square is uninteresting
			  break;

			case waterChar:
			  square |= SBUtils::flag_lethal_neighbour;
			  break;

			case baseChar:
			case freeChar:
			  square |= TinyEdgeSet::mask_w;
			  temp.mBoardBits[rawOffset - 1]
			    |= TinyEdgeSet::mask_e;

			  check (temp.mBoardBits[rawOffset - 1]);
			  break;
			}
		    }

		  switch (prevRow[column])
		    {
		    case wallChar:
		      break;

		    case waterChar:
		      square |= SBUtils::flag_lethal_neighbour;
		      break;

		    case baseChar:
		    case freeChar:
		      square |= TinyEdgeSet::mask_s;
		      temp.mBoardBits[rawOffset - temp.mWidth]
			|= TinyEdgeSet::mask_n;

		      check (temp.mBoardBits[rawOffset - temp.mWidth]);
		      break;
		    }
		  break;
		}

	      temp.mBoardBits[rawOffset++] = square;
	    }

	  std::swap (prevRow, thisRow);
	}
    }

  if (strm)
    {
      swap (temp);
    }
}

/////////////////////////////////////////////////////////////////////////
// Extraction operator
/////////////////////////////////////////////////////////////////////////

std::istream &operator>> (std::istream &strm, StaticBoard &board)
{
  board.load (strm);
  return strm;
}

/////////////////////////////////////////////////////////////////////////
// Debugging dump
/////////////////////////////////////////////////////////////////////////

std::ostream &operator<< (std::ostream &strm, StaticBoard const &board)
{
  for (int y = board.height() - 1; y >= 0; --y)
    {
      for (int x = 0; x < board.width(); ++x)
	{
	  strm << board (x, y);
	}

      strm << "\n";
    }

  return strm;
}

/////////////////////////////////////////////////////////////////////////////
// Find out lethal edges
/////////////////////////////////////////////////////////////////////////////

TinyEdgeSet StaticBoard::lethalEdges (int offs) const
{
  TinyEdgeSet result (0);

  BitSet square = rawBits (offs);

  if (SBUtils::hasLethalNeighbour (square))
    {
      if ((square & TinyEdgeSet::north) && (isLethal (offs + mWidth)))
	{
	  result.addEdge (TinyEdgeSet::north);
	}

      if ((square & TinyEdgeSet::south) && (isLethal (offs - mWidth)))
	{
	  result.addEdge (TinyEdgeSet::south);
	}

      if ((square & TinyEdgeSet::east) && (isLethal (offs + 1)))
	{
	  result.addEdge (TinyEdgeSet::east);
	}

      if ((square & TinyEdgeSet::west) && (isLethal (offs - 1)))
	{
	  result.addEdge (TinyEdgeSet::west);
	}
    }

  return result;
}
