// -*- mode:c++ -*-
//
// Header file StaticBoard.hh
//
// Copyright (c) 2002 Raoul M. Gough
//
// Represents the static portions of the game board (walls, water and
// home bases)
//
// History
// =======
// 2002/ 8/31	rmg	File creation
//
// $Id: StaticBoard.hh,v 1.10 2002/09/02 18:23:49 raoul Exp $
//

#ifndef StaticBoard_rmg_20020831_included
#define StaticBoard_rmg_20020831_included

#include <iosfwd>
#include <vector>
#include <algorithm>

#include "TinyEdgeSet.hh"
#include "TESGraph.hh"
#include "Location.hh"

class StaticBoard
{
  friend class SBUtils;

public:
  StaticBoard () : mWidth (0), mHeight (0) { }

  void load (std::istream &);

  int width () const { return mWidth; }
  int height () const { return mHeight; }

  //
  // Coordinates are (x, y) ranging from
  //
  // x: [0..width()-1]
  // y: [0..height()-1]
  //
  // Thus the southwest corner is (0, 0) and
  // northeast is (width()-1, height()-1)
  //

  char operator() (int, int) const; // Return one of .#@~
  // Note: this is *slower* than calling any two of the next five functions

  int offset (int x, int y) const { return y * mWidth + x; }
  int offset (Location const &loc) const {
    return loc.first + mWidth * loc.second;
  }

  Location whereIs (int offset) const {
    return Location (offset % mWidth, offset / mWidth);
  }

  inline bool isBase             (int offset) const;
  inline bool isWall             (int offset) const;
  inline bool isLethal           (int offset) const;
  inline bool hasLethalNeighbour (int offset) const;

  inline TinyEdgeSet edges       (int offset) const;

  TinyEdgeSet lethalEdges (int offset) const;

  void swap (StaticBoard &other)
  {
    mBoardBits.swap (other.mBoardBits);
    std::swap (mWidth, other.mWidth);
    std::swap (mHeight, other.mHeight);
  }

  TESGraph makeGraph () const { return TESGraph (&mBoardBits, mWidth); }

  std::vector<int> const &bases () const { return mBases; }

private:
  typedef unsigned char BitSet;
  typedef std::vector<BitSet> BoardType;

  BitSet rawBits (int offs) const { return mBoardBits[offs]; }

  int mWidth;
  int mHeight;
  BoardType mBoardBits;
  std::vector<int> mBases;
};

std::istream &operator>> (std::istream &, StaticBoard &);
std::ostream &operator<< (std::ostream &, StaticBoard const &);

struct SBUtils
{
  enum {
    flag_base = 1 << TinyEdgeSet::dir_limit
    , flag_lethal = flag_base << 1
    , flag_lethal_neighbour = flag_lethal << 1
    , flag_wall = flag_lethal_neighbour << 1
    , flag_all_edges = flag_base - 1
  };

  static bool isBase (StaticBoard::BitSet set)
  {
    return set & flag_base;
  }

  static bool isWall (StaticBoard::BitSet set)
  {
    return !(set & (flag_lethal | flag_base | flag_all_edges));
  }

  static bool isLethal (StaticBoard::BitSet set)
  {
    return set & flag_lethal;
  }

  static bool hasLethalNeighbour (StaticBoard::BitSet set)
  {
    return set & flag_lethal_neighbour;
  }
};

bool StaticBoard::isBase (int offs) const
{
  return SBUtils::isBase (rawBits (offs));
}

bool StaticBoard::isWall (int offs) const
{
  return SBUtils::isWall (rawBits (offs));
}

bool StaticBoard::isLethal (int offs) const
{
  return SBUtils::isLethal (rawBits (offs));
}

bool StaticBoard::hasLethalNeighbour (int offs) const
{
  return SBUtils::hasLethalNeighbour (rawBits (offs));
}

TinyEdgeSet StaticBoard::edges (int offs) const
{
  return TinyEdgeSet (rawBits (offs) & SBUtils::flag_all_edges);
}

inline bool isBase (StaticBoard const &board, int x, int y)
{
  return board.isBase (board.offset (x, y));
}

inline bool isWall (StaticBoard const &board, int x, int y)
{
  return board.isWall (board.offset (x, y));
}

inline bool isLethal (StaticBoard const &board, int x, int y)
{
  return board.isLethal (board.offset (x, y));
}

inline bool hasLethalNeighbour (StaticBoard const &board, int x, int y)
{
  return board.hasLethalNeighbour (board.offset (x, y));
}

inline TinyEdgeSet edges (StaticBoard const &board, int x, int y)
{
  return board.edges (board.offset (x, y));
}

inline TinyEdgeSet lethalEdges (StaticBoard const &board, int x, int y)
{
  return board.lethalEdges (board.offset (x, y));
}

#endif // StaticBoard_rmg_20020831_included
