// -*- mode:c++ -*-
//
// Header file TESGraph.hh
//
// Boost graph adapter for TinyEdgeSet
//
// Copyright (c) 2002 Raoul M. Gough
//
// History
// =======
// 2002/ 9/ 1	rmg	File creation
//
// $Id: TESGraph.hh,v 1.4 2002/09/01 21:04:34 raoul Exp $
//

#ifndef TESGraph_rmg_20020901_included
#define TESGraph_rmg_20020901_included

#include "TinyEdgeSet.hh"
#include "TinyEdgeIterator.hh"
#include "TinyEdgeDestIterator.hh"

#include <boost/graph/graph_concepts.hpp>
#include <boost/pending/integer_range.hpp>
#include <vector>
#include <utility>

struct TESGraph_graph_tag;

class TESGraph
{
public:
  TESGraph (std::vector<unsigned char> const *sbptr, int width)
    : mSetPtr (sbptr)
    , mWidth (width)
  {
  }

  inline TinyEdgeIterator makeIterator (int v) const;
  inline TinyEdgeDestIterator makeDestIterator (int v) const;

  TinyEdgeSet operator[] (int n) const
  {
    return reinterpret_cast<TinyEdgeSet const &>((*mSetPtr)[0]);
  }

  size_t vertices () const
  {
    debugStream() << " TG" << mSetPtr->size() << " ";

    return mSetPtr->size();
  }

  typedef boost::no_property vertex_property_type;
  typedef TESGraph_graph_tag graph_tag;

private:
  std::vector<unsigned char> const *mSetPtr;
  int mWidth;
};

struct TESGraph_traversal_tag
  : public virtual boost::incidence_graph_tag
  , public virtual boost::adjacency_graph_tag
  , public virtual boost::vertex_list_graph_tag { };

// public virtual boost::edge_list_graph_tag
// public virtual boost::adjacency_matrix_tag

namespace boost
{
  template<>
  struct graph_traits<TESGraph>
  {
    typedef int V;
    typedef V vertex_descriptor;

    typedef TinyEdgeSet::Edge edge_descriptor;
    typedef TinyEdgeDestIterator adjacency_iterator;
    typedef TinyEdgeIterator out_edge_iterator;
    typedef integer_range<V>::iterator vertex_iterator;
    typedef undirected_tag directed_category;
    typedef disallow_parallel_edge_tag edge_parallel_category;
    typedef TESGraph_traversal_tag traversal_category;
    typedef size_t vertices_size_type;
    typedef void edges_size_type;
    typedef size_t degree_size_type;
    typedef TinyEdgeIterator edge_iterator;
  };
}

////////////////////////////////////////////////////////////////////////////
// Make an iterator for the edges of the current vertex
////////////////////////////////////////////////////////////////////////////

TinyEdgeIterator TESGraph::makeIterator (int n) const
{
  TinyEdgeSet const *base
    = reinterpret_cast<TinyEdgeSet const *>(&(*mSetPtr)[0]);

  return
    TinyEdgeIterator (base + n
		      , base + mSetPtr->size()
		      , n
		      , mWidth);
}

////////////////////////////////////////////////////////////////////////////
// Make an iterator for the adjacent vertices of the current vertex 
////////////////////////////////////////////////////////////////////////////

TinyEdgeDestIterator TESGraph::makeDestIterator (int n) const
{
  TinyEdgeSet const *base
    = reinterpret_cast<TinyEdgeSet const *>(&(*mSetPtr)[0]);

  return
    TinyEdgeDestIterator (base + n
			  , base + mSetPtr->size()
			  , n
			  , mWidth);
}

////////////////////////////////////////////////////////////////////////////
// IncidenceGraph support
////////////////////////////////////////////////////////////////////////////

inline
std::pair<boost::graph_traits<TESGraph>::out_edge_iterator
         , boost::graph_traits<TESGraph>::out_edge_iterator>
out_edges (int node, TESGraph const &g)
{
  return std::make_pair (g.makeIterator (node), g.makeIterator (node + 1));
}

inline int source (TinyEdgeSet::Edge const &edge, TESGraph const &)
{
  return edge.first;
}

inline int target (TinyEdgeSet::Edge const &edge, TESGraph const &)
{
  return edge.second;
}

inline size_t out_degree (int v, TESGraph const &graph)
{
  return graph[v].edgeCount();
}

////////////////////////////////////////////////////////////////////////////
// AdjacencyGraph support
////////////////////////////////////////////////////////////////////////////

inline
std::pair<boost::graph_traits<TESGraph>::adjacency_iterator
         , boost::graph_traits<TESGraph>::adjacency_iterator>
adjacent_vertices (int node, TESGraph const &g)
{
  return std::make_pair (g.makeDestIterator (node)
			 , g.makeDestIterator (node + 1));
}

////////////////////////////////////////////////////////////////////////////
// AdjacencyGraph support
////////////////////////////////////////////////////////////////////////////

inline
std::pair<boost::integer_range<int>::iterator
          , boost::integer_range<int>::iterator>
vertices(TESGraph const &graph)
{
  debugStream() << "TESGraph vertices";

  return std::make_pair
    (boost::integer_range<int>::iterator (0)
     , boost::integer_range<int>::iterator (graph.vertices()));
}

inline size_t num_vertices (TESGraph const &graph) { return graph.vertices(); }

////////////////////////////////////////////////////////////////////////////
// Partial EdgeListGraph
////////////////////////////////////////////////////////////////////////////

inline
std::pair<boost::graph_traits<TESGraph>::edge_iterator
         , boost::graph_traits<TESGraph>::edge_iterator>
edges (TESGraph const &g)
{
  debugStream() << "TESGraph edges to";

  return std::make_pair (g.makeIterator (0)
			 , g.makeIterator (g.vertices()));
}

#endif // TESGraph_rmg_20020901_included
