// -*- mode:c++ -*-
//
// Header file TinyEdgeSet.hh
//
// boost::graph compatible edge list which stores edges from the set
// N, S, E, W.
//
// Copyright (c) 2002 Raoul M. Gough
//
// History
// =======
// 2002/ 9/ 1	rmg	File creation
//
// $Id: TinyEdgeSet.hh,v 1.3 2002/09/01 16:46:29 raoul Exp $
//

#ifndef TinyEdgeSet_rmg_20020901_included
#define TinyEdgeSet_rmg_20020901_included

#include <cassert>

class TinyEdgeSet
{
  friend class TinyEdgeIterator;

public:
  typedef enum
  {
    north = 0
    , south
    , east
    , west
    , dir_limit
    , dir_base = 0
  } Direction;

  TinyEdgeSet (int bits = 0) : mBits (bits & mask_all) { }

  void addEdge    (Direction d)       { mBits |= (1 << d); }
  void removeEdge (Direction d)       { mBits &= ~(1 << d); }
  bool hasEdge    (Direction d) const { return mBits & (1 << d); }

  // Low-level interface
  void bits       (int bits)       { mBits = bits; }
  int  bits       ()         const { return mBits; }

  void addBits    (int bits)       { mBits |= bits; }
  void removeBits (int bits)       { mBits &= ~bits; }

  bool hasAllBits (int bits) const { return bits == (mBits & bits); }
  bool hasAnyBits (int bits) const { return mBits & bits; }

  inline int edgeCount () const;

  enum {
    mask_n = 1 << north
    , mask_s = 1 << south
    , mask_e = 1 << east
    , mask_w = 1 << west
    , mask_all = mask_n | mask_s | mask_e | mask_w
  };

public:
  //
  // Boost graph library adaptation
  //

  typedef int value_type;

  struct Edge : public std::pair<int, int>
  {
    static int offset (int from, int width, Direction d)
    {
      switch (d)
	{
	case north:
	  return from + width;

	case south:
	  return from - width;

	case east:
	  return from + 1;

	case west:
	  return from - 1;

	default:
	  assert (false);
	}
    }

    Edge (int from, int width, Direction dir)
      : std::pair<int, int> (from, offset (from, width, dir))
    {
    }

    Edge () { }
  };

private:
  char mBits;
};

int TinyEdgeSet::edgeCount () const
{
  int count = 0;

  for (int bit = mask_n; bit < mask_all; bit <<= 1)
    {
      if (mBits & bit)
	{
	  ++count;
	}
    }

  return count;
}

#endif // TinyEdgeSet_rmg_20020901_included
