// -*- mode:c++ -*-
//
// Module World.cc
//
// Copyright (c) 2002 Raoul M. Gough
//
// History
// =======
// 2002/ 8/31	rmg	File creation
//
// $Id: World.cc,v 1.8 2002/09/02 18:23:23 raoul Exp raoul $
//

#include "World.hh"

#include <iostream>
#include <string>
#include <sstream>
#include <limits.h>
#include <algorithm>
#include <vector>

#include "LogStream.hh"
#include "RobotInfo.hh"

/////////////////////////////////////////////////////////////////////////////
// Constructor
/////////////////////////////////////////////////////////////////////////////

World::World (std::istream &strm)
  : mStaticBoard ()
{
  strm >> mStaticBoard;

  RobotInfo myRobotInfo;

  strm >> myRobotInfo;

  mRobotPtrs.push_back
    (boost::shared_ptr<Robot> (new SmartRobot (*this, myRobotInfo)));

  strm.ignore (INT_MAX, '\n');  // Skip end of line

  update (strm);

  mSearcherPtr = boost::shared_ptr<Searcher> (new Searcher (*this));
}

/////////////////////////////////////////////////////////////////////////////
// Lookup square info
/////////////////////////////////////////////////////////////////////////////

SquareInfo const *World::knownInfo (int loc) const
{
  std::map<int, SquareInfo>::const_iterator iter = mLoadedSquares.find (loc);
      
  if (iter == mLoadedSquares.end())
    {
      return 0;
    }

  else
    {
      return &iter->second;
    }
}

SquareInfo *World::knownInfo (int loc)
{
  return const_cast <SquareInfo *>(const_cast<World const *>(this)->knownInfo(loc));
}

/////////////////////////////////////////////////////////////////////////////
// Read a robot update line
/////////////////////////////////////////////////////////////////////////////

void World::update (std::istream &rawStream)
{
  readRobots (rawStream);
  readPackages (rawStream);
}

/////////////////////////////////////////////////////////////////////////////
// Read a package update line
/////////////////////////////////////////////////////////////////////////////

void World::readPackages (std::istream &rawStream)
{
  std::string packageLine;

  getline (rawStream, packageLine);

  std::istringstream packageStream (packageLine);

  PackageInfo package;

  packageStream >> package;

  commStream() << "Server: " << packageLine << " (length " << packageLine.length() << ")" << std::endl;

  int location = smartRobot()->location();

  if (packageLine.length() && packageStream.good())
    {
      SquareInfo &square = mLoadedSquares[smartRobot()->location()];
  
      square.packages().reset ();

      do
	{
	  square.packages().insert (package);
	}
      while (packageStream >> package);

      logStream()
	<< "Square at "
	<< smartRobot()->location()
	<< " has "
	<< square.packages().size()
	<< " packages with total weight "
	<< square.packages().totalWeight()
	<< "\n";
    }

  else
    {
      // No packages here. Synchronize record

      SquareInfo *squarePtr = knownInfo (location);
      
      if (squarePtr)
	{
	  if (squarePtr->packages().size() != 0)
	    {
	      logStream()
		<< "Square at "
		<< smartRobot()->location()
		<< " is now empty!\n";
	    }

	  squarePtr->packages().reset ();
	}
    }
}

/////////////////////////////////////////////////////////////////////////////
// Read a robot update line
/////////////////////////////////////////////////////////////////////////////

void World::readRobots (std::istream &rawStream)
{
  std::string robotLine;

  getline (rawStream, robotLine);

  commStream() << "Server: " << robotLine << std::endl;

  std::istringstream robotStream (robotLine);

  std::vector<int> livingRobots;

  while (robotStream)
    {
      char hash;

      robotStream >> hash;

      if (robotStream && (hash != '#'))
	{
	  robotStream.setstate (std::ios::failbit);
	}

      else
	{
	  int robotId;

	  robotStream >> robotId;

	  std::string temp;

	  getline (robotStream, temp, '#');

	  if (robotStream.good() || robotStream.eof())
	    {
	      if (!robotStream.eof())
		{
		  robotStream.unget(); // Put back the hash!
		}

	      livingRobots.push_back (robotId); // Robot is alive

	      std::istringstream commandStream (temp);

	      readRobot (robotId, commandStream);
	    }
	}
    }

  std::sort (livingRobots.begin(), livingRobots.end());

  RobotsType::iterator iter = mRobotPtrs.begin();

  while (iter != mRobotPtrs.end())
    {
      if (std::binary_search (livingRobots.begin()
			      , livingRobots.end()
			      , (*iter)->id()))
	{
	  ++iter;
	}

      else
	{
	  logStream()
	    << "Robot "
	    << (*iter)->id()
	    << " has died with score approx. "
	    << (*iter)->score()
	    << "\n";

	  iter = mRobotPtrs.erase (iter);
	}
    }
}

/////////////////////////////////////////////////////////////////////////////
// Read commands for a single robot
/////////////////////////////////////////////////////////////////////////////

void World::readRobot (int robotId, std::istream &commandStream)
{
  char command;

  while (commandStream >> command)
    {
      switch (command)
	{
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 'X':
	  {
	    commandStream.unget();

	    Location loc;
	    commandStream >> loc;
	    addRobot (robotId, loc);
	  }
	  break;

	case 'N':
	  moveRobot (robotId, 0, 1);
	  break;

	case 'S':
	  moveRobot (robotId, 0, -1);
	  break;

	case 'E':
	  moveRobot (robotId, 1, 0);
	  break;

	case 'W':
	  moveRobot (robotId, -1, 0);
	  break;

	case 'P':
	  {
	    int packageId;
	    commandStream >> packageId;
	    pickPackage (robotId, packageId);
	  }
	  break;

	case 'D':
	  {
	    int packageId;
	    commandStream >> packageId;
	    dropPackage (robotId, packageId);
	  }
	  break;

	default:
	  logStream()
	    << "warning: Unrecognised robot command "
	    << command
	    << std::endl;

	  // Now ignore the error and hope for the best
	  break;
	}
    }
}

/////////////////////////////////////////////////////////////////////////////
// Find robot
/////////////////////////////////////////////////////////////////////////////

Robot *World::robot (int id) const
{
  // FIXME (yeah, right). Use a better data structure.

  for (RobotsType::const_iterator iter = mRobotPtrs.begin()
	 ; iter != mRobotPtrs.end()
	 ; ++iter)
    {
      if ((*iter)->id() == id)
	{
	  return iter->get();
	}
    }

  return 0;
}

/////////////////////////////////////////////////////////////////////////////
// New robot
/////////////////////////////////////////////////////////////////////////////

void World::addRobot (int robotId, Location const &loc)
{
  if (robotId == mRobotPtrs.front()->id())
    {
      mRobotPtrs.front()->setLocation (mStaticBoard.offset (loc));

      logStream()
	<< "I am "
	<< mRobotPtrs.front()->id()
	<< " at "
	<< mRobotPtrs.front()->location()
	<< " "
	<< mStaticBoard.whereIs (mRobotPtrs.front()->location())
	<< " "
	<< loc
	<< " capacity "
	<< mRobotPtrs.front()->capacity()
	<< " money "
	<< mRobotPtrs.front()->money()
	<< "\n";
    }

  else
    {
      if (!robot (robotId))
	{
	  mRobotPtrs.push_back
	    (boost::shared_ptr<Robot> (new Robot (*this, robotId)));

	  mRobotPtrs.back()->setLocation (mStaticBoard.offset (loc));

	  logStream()
	    << "New robot "
	    << mRobotPtrs.back()->id()
	    << " at "
	    << loc
	    << "\n";
	}
    }
}

/////////////////////////////////////////////////////////////////////////////
// Robot move
/////////////////////////////////////////////////////////////////////////////

void World::moveRobot (int robotId, int x, int y)
{
  Robot *robotPtr = robot (robotId);

  robotPtr->setLocation (robotPtr->location() + mStaticBoard.offset (x, y));

  logStream()
    << "Robot "
    << robotId
    << " moved "
    << Location (x, y)
    << " to "
    << mStaticBoard.whereIs (robotPtr->location ())
    << "\n";
}

/////////////////////////////////////////////////////////////////////////////
// Robot picked up a package
/////////////////////////////////////////////////////////////////////////////

void World::pickPackage (int robotId, int packageId)
{
  Robot *robotPtr = robot (robotId);

  SquareInfo *squarePtr = knownInfo (robotPtr->location());

  PackageInfo package (packageId);

  if (squarePtr)
    {
      package = squarePtr->packages().remove (packageId);
    }

  robotPtr->pickPackage (package);

  logStream()
    << "Robot "
    << robotId
    << " picked up package "
    << package
    << "\n";
}

/////////////////////////////////////////////////////////////////////////////
// Robot dropped a package
/////////////////////////////////////////////////////////////////////////////

void World::dropPackage (int robotId, int packageId)
{
  Robot *robotPtr = robot (robotId);

  PackageInfo package = robotPtr->dropPackage (packageId);

  logStream()
    << "Robot "
    << robotId
    << " dropped package "
    << package;

  if (mStaticBoard.offset (package.mDestX, package.mDestY)
      == robotPtr->location())
    {
      // Package disappears

      logStream() << " and scored!";
    }

  else
    {
      mLoadedSquares[robotPtr->location()].packages().insert (package);
    }

  logStream()
    << "\n";
}

/////////////////////////////////////////////////////////////////////////////
// Get a spanning tree
/////////////////////////////////////////////////////////////////////////////

boost::shared_ptr<SpanningTree const> World::tree (int location)
{
  SquareInfo &square = mLoadedSquares[location];

  boost::shared_ptr<SpanningTree> result = square.tree();

  if (!result.get())
    {
      result
	= boost::shared_ptr<SpanningTree>
	(new SpanningTree (searcher().treeSize(), location));

      searcher().search (*result);

      // Save this for next time
      square.tree (result);
    }

  return result;
}

/////////////////////////////////////////////////////////////////////////////
// Get the set of package groups and unexplored bases
/////////////////////////////////////////////////////////////////////////////

std::vector<AddressInfo> World::packagesAndBases () const
{
  std::vector<AddressInfo> result;

  for (std::vector<int>::const_iterator iter = mStaticBoard.bases().begin()
	 ; iter != mStaticBoard.bases().end()
	 ; ++iter)
    {
      SquareInfo const *known = knownInfo (*iter);

      if ((!known) || (!known->packages().size()))
	{
	  AddressInfo temp (*iter);

	  // Should adapt these dynamically
	  temp.mWeight = 1000;
	  temp.mPackages = 1000;

	  result.push_back (temp);
	}
    }

  for (std::map<int, SquareInfo>::const_iterator iter = mLoadedSquares.begin()
	 ; iter != mLoadedSquares.end()
	 ; ++iter)
    {
      if (iter->second.packages().size())
	{
	  AddressInfo temp (iter->first);

	  temp.mWeight = iter->second.packages().totalWeight();
	  temp.mPackages = iter->second.packages().size();
	}
    }

  return result;
}
