// -*- mode:c++ -*-
//
// Header file World.hh
//
// Copyright (c) 2002 Raoul M. Gough
//
// A model of the playing world
//
//                   +-----L-+
//                   | robot |
//                   +-+---+-+
//             	      /     \
//          _________N       \
//         /   	              \
//        /    	       	       \
// +---+-+-+       +-----L-+    |
// | world +-----N-+ base  |    |
// +---+---+       +-------+    |
//     |       	       	        |
//     |       	       	        |
//     |               	       /
//     N               	      N
//     |       	       	     /
// +---+--L-+       +-------+-+
// | square |-----N-+ package |
// +--------+       +---------+
//
// N = Cardinality 0..n
// 1 = Cardinality 1
// L = object has a location
//
// History
// =======
// 2002/ 8/31	rmg	File creation
//
// $Id: World.hh,v 1.7 2002/09/02 18:23:29 raoul Exp $
//

#ifndef World_rmg_20020831_included
#define World_rmg_20020831_included

#include "StaticBoard.hh"
#include "Location.hh"
#include "Robot.hh"
#include "SmartRobot.hh"
#include "Searcher.hh"
#include "SquareInfo.hh"

#include <iosfwd>
#include <list>
#include <boost/smart_ptr.hpp>

class World
{
public:
  World (std::istream &); // Read initial details from stream

  void update (std::istream &); // Read server response

  StaticBoard const &board () const { return mStaticBoard; }

  SmartRobot *smartRobot() const
  {
    return dynamic_cast<SmartRobot *>(mRobotPtrs.front().get());
  }

  Robot *robot (int id) const;

  Searcher &searcher () const { return *mSearcherPtr; }

  SquareInfo const *knownInfo (int) const;
  SquareInfo *knownInfo (int);

  boost::shared_ptr<SpanningTree const> tree (int location);

  std::vector<AddressInfo> packagesAndBases () const;

private:
  void readRobots (std::istream &); // Process a full robot command line
  void readRobot (int id, std::istream &); // Process a robot's commands

  void readPackages (std::istream &); // Process a full package list

  void addRobot (int id, Location const &);
  void moveRobot (int id, int x, int y);
  void pickPackage (int robot, int package);
  void dropPackage (int robot, int package);

private:
  StaticBoard mStaticBoard;
  int mMyId;
  typedef std::list<boost::shared_ptr<Robot> > RobotsType;
  RobotsType mRobotPtrs;

  boost::shared_ptr<Searcher> mSearcherPtr;
  std::map<int, SquareInfo> mLoadedSquares;
};

#endif // World_rmg_20020831_included
